# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import String, HashMap, U32


class TestHashmap(unittest.TestCase):

    def test_hashmap_encode(self):
        obj = HashMap(String, U32).new()
        data = obj.encode([('1', 2), ('23', 24), ('28', 30), ('45', 80)])
        self.assertEqual(data.to_hex(), '0x10043102000000083233180000000832381e00000008343550000000')

    def test_hashmap_decode(self):
        obj = HashMap(String, U32).new()
        data = ScaleBytes("0x10043102000000083233180000000832381e00000008343550000000")
        self.assertEqual([('1', 2), ('23', 24), ('28', 30), ('45', 80)], obj.decode(data))


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import F64, F32


class TestFloat(unittest.TestCase):

    def test_f64_decode(self):
        obj = F64.new()
        obj.decode(ScaleBytes("0x333333333333f33f"))
        self.assertAlmostEqual(obj.value, 1.2)

    def test_f32_decode(self):
        obj = F32.new()
        obj.decode(ScaleBytes("0x9a99993f"))
        self.assertAlmostEqual(obj.value, 1.2)

    def test_f64_encode(self):
        obj = F64.new()
        obj.encode(-0.0)
        self.assertEqual(str(obj.data), "0x0000000000000080")

    def test_f64_encode_invalid_input(self):
        obj = F64.new()
        with self.assertRaises(ValueError) as cm:
            obj.encode(-0)
            self.assertEqual('0 is not a float', str(cm.exception))

    def test_f32_encode(self):
        obj = F32.new()
        obj.encode(-0.0)
        self.assertEqual(str(obj.data), "0x00000080")


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import Optional, Union

from scalecodec.constants import TYPE_DECOMP_MAX_RECURSIVE
from scalecodec.exceptions import RemainingScaleBytesNotEmptyException, ScaleDecodeException


class ScaleBytes:
    """
    Representation of SCALE encoded Bytes.
    """

    def __init__(self, data: Union[str, bytes, bytearray, int]):
        """
        Constructs a SCALE bytes-stream with provided `data`

        Parameters
        ----------
        data
        """
        self.offset = 0

        if type(data) is bytearray:
            self.data = data
        elif type(data) is bytes:
            self.data = bytearray(data)
        elif type(data) is str and data[0:2] == '0x':
            self.data = bytearray.fromhex(data[2:])
        else:
            raise ValueError("Provided data is not in supported format: provided '{}'".format(type(data)))

        self.length = len(self.data)

    def get_next_bytes(self, length: int) -> bytearray:
        """
        Retrieve `length` amount of bytes of the stream

        Parameters
        ----------
        length: amount of requested bytes

        Returns
        -------
        bytearray
        """
        if self.offset + length > self.length:
            raise RemainingScaleBytesNotEmptyException(
                f'No more bytes available (needed: {self.offset + length} / total: {self.length})'
            )

        data = self.data[self.offset:self.offset + length]
        self.offset += length
        return data

    def get_remaining_bytes(self) -> bytearray:
        """
        Retrieves all remaining bytes from the stream

        Returns
        -------
        bytearray
        """

        data = self.data[self.offset:]
        self.offset = self.length
        return data

    def get_remaining_length(self) -> int:
        """
        Returns how many bytes are left in the stream

        Returns
        -------
        int
        """
        return self.length - self.offset

    def reset(self):
        """
        Resets the pointer of the stream to the beginning

        Returns
        -------

        """
        self.offset = 0

    def copy(self):
        return ScaleBytes(self.data)

    def __str__(self):
        return "0x{}".format(self.data.hex())

    def __eq__(self, other):
        if not hasattr(other, 'data'):
            return False
        return self.data == other.data

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return "<{}(data=0x{})>".format(self.__class__.__name__, self.data.hex())

    def __add__(self, data):

        if type(data) is ScaleBytes:
            return ScaleBytes(self.data + data.data)

        if type(data) is bytes:
            data = bytearray(data)
        elif type(data) == str and data[0:2] == '0x':
            data = bytearray.fromhex(data[2:])

        if type(data) is bytearray:
            return ScaleBytes(self.data + data)

    def __bytes__(self):
        return self.to_bytes()

    def to_bytes(self):
        return bytes(self.data)

    def to_hex(self) -> str:
        """
        Return a hex-string (e.g. "0x00") representation of the byte-stream

        Returns
        -------
        str
        """
        return f'0x{self.data.hex()}'


class ScaleTypeDef:

    scale_type_cls = None

    def __init__(self, name: str = None, metadata=None):
        if self.scale_type_cls is None:
            self.scale_type_cls = ScaleType
        self.name = name
        self.runtime_config = None
        self.metadata = metadata

    def new(self, **kwargs) -> 'ScaleType':
        obj = self.scale_type_cls(type_def=self, **kwargs)
        if 'value' in kwargs:
            obj.deserialize(kwargs['value'])

        if 'scale' in kwargs:
            obj.decode(kwargs['scale'])

        return obj

    def impl(self, scale_type_cls: type = None, runtime_config=None) -> 'ScaleTypeDef':
        """

        Returns:
            object:
        """
        if scale_type_cls:
            self.scale_type_cls = scale_type_cls
        if runtime_config:
            self.runtime_config = runtime_config

        return self

    # def create_from_registry_type(self, registry_type):

    @abstractmethod
    def _encode(self, value: any) -> ScaleBytes:
        pass

    def encode(self, value: any, external_call=True) -> ScaleBytes:

        if external_call:
            raise ValueError("encode of definition cannot be called directly")
        #
        # if issubclass(value.__class__, ScaleType):
        #     if value.type_def.__class__ is self.__class__:
        #         return value.data
        #     else:
        #         raise ValueError(f"Cannot encode '{value.type_def.__class__}' to a '{self.__class__}'")
        # else:
        return self._encode(value)

    @abstractmethod
    def decode(self, data: ScaleBytes) -> any:
        pass

    @abstractmethod
    def serialize(self, value: any) -> any:
        raise NotImplementedError()

    @abstractmethod
    def deserialize(self, value: any) -> any:
        raise NotImplementedError()

    # TODO implement

    @abstractmethod
    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):

        if _recursion_level > max_recursion:
            return self.__class__.__name__
        return self.__class__.__name__


class ScaleType:

    def __init__(self, type_def: ScaleTypeDef, **kwargs):
        # TODO remove kwargs

        self.meta_info = None
        self.type_def: ScaleTypeDef = type_def
        self.value_serialized = None
        self.value_object = None

        self._data = None
        self._data_start_offset = 0
        self._data_end_offset = 0

        super().__init__()

    # def __call__(self, *args, **kwargs):
    #     return self

    def encode(self, value: Optional[any] = None) -> ScaleBytes:
        if value is not None and issubclass(self.__class__, value.__class__):
            # Accept instance of current class directly
            self._data = value.data
            self.value_object = value.value_object
            self.value_serialized = value.value_serialized
            return value.data

        if value is None:
            value = self.value_serialized

        self._data = self.type_def.encode(value, False)
        self._data_start_offset = self._data.offset
        self._data_end_offset = self._data.length

        self.value_serialized = value
        self.value_object = self.deserialize(value)

        return self._data

    def decode(self, data: ScaleBytes, check_remaining=False) -> any:
        self._data = data
        self._data_start_offset = data.offset
        # Decode type
        self.value_object = self.type_def.decode(data)

        self._data_end_offset = data.offset

        if check_remaining and self._data_end_offset != self._data.length:
            raise ScaleDecodeException(
                f'Remaining ScaleBytes - Current offset: {self._data_end_offset} / length: {self._data.length}'
            )

        self.value_serialized = self.serialize()
        return self.value_serialized

    def serialize(self) -> Union[int, str, dict, tuple, bool]:
        self.value_serialized = self.type_def.serialize(self.value_object)
        return self.value_serialized

    def deserialize(self, value_serialized: any):
        if value_serialized and issubclass(self.__class__, self.value_serialized.__class__):
            # Accept instance of current class directly
            self.value_object = self.value_serialized.value_object
            self.value_serialized = self.value_serialized.value_serialized
            return self.value_object

        self.value_object = self.type_def.deserialize(value_serialized)
        self.value_serialized = self.type_def.serialize(self.value_object)

        return self.value_object

    @property
    def value(self):
        return self.value_serialized

    @value.setter
    def value(self, value):
        self.value_serialized = value

    @property
    def data(self) -> Optional[ScaleBytes]:
        """
        Returns a ScaleBytes instance of the SCALE-bytes used in the decoding process

        Returns
        -------
        bytearray
        """
        if self._data is not None:
            return ScaleBytes(self._data.data[self._data_start_offset:self._data_end_offset])

    def example_value(self):
        return self.type_def.example_value()

    def __repr__(self):

        # if self.__class__ is not ScaleType:
        #     name = self.__class__.__name__
        # else:
        #     name = self.type_def.__class__.__name__

        name = self.type_def.__class__.__name__

        if self.value_serialized is not None:
            return f"<{name}(value={self.value_serialized})>"
        elif self.data:
            return f"<{name}(data={self.data.to_hex()})>"
        else:
            return f"<{name}>"

    def __getitem__(self, item):
        return self.value_object[item]

    def __iter__(self):
        for item in self.value_object:
            yield item

    def __eq__(self, other):
        if isinstance(other, ScaleType):
            return other.value_serialized == self.value_serialized
        else:
            return other == self.value_serialized

    def __gt__(self, other):
        if isinstance(other, ScaleType):
            return self.value_serialized > other.value_serialized
        else:
            return self.value_serialized > other

    def __ge__(self, other):
        if isinstance(other, ScaleType):
            return self.value_serialized >= other.value_serialized
        else:
            return self.value_serialized >= other

    def __lt__(self, other):
        if isinstance(other, ScaleType):
            return self.value_serialized < other.value_serialized
        else:
            return self.value_serialized < other

    def __le__(self, other):
        if isinstance(other, ScaleType):
            return self.value_serialized <= other.value_serialized
        else:
            return self.value_serialized <= other


class ScalePrimitive(ScaleTypeDef):
    pass



# Python SCALE Codec Library
#
# Copyright 2018-2023 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import enum

import math
import struct
from typing import Union, Optional

from scalecodec.base import ScaleType, ScaleBytes, ScalePrimitive, ScaleTypeDef
from scalecodec.constants import TYPE_DECOMP_MAX_RECURSIVE
from scalecodec.exceptions import ScaleEncodeException, ScaleDecodeException, ScaleDeserializeException, \
    ScaleSerializeException


class UnsignedInteger(ScalePrimitive):
    """
    Unsigned int type, encoded in little-endian (LE) format
    """

    def __init__(self, bits: int):
        super().__init__()
        self.bits = bits
        self.byte_count = int(self.bits / 8)

    def decode(self, data: ScaleBytes) -> int:
        return int.from_bytes(data.get_next_bytes(self.byte_count), byteorder='little')

    def _encode(self, value) -> ScaleBytes:

        if 0 <= int(value) <= 2**(self.byte_count * 8) - 1:
            return ScaleBytes(bytearray(int(value).to_bytes(self.byte_count, 'little')))
        else:
            raise ScaleEncodeException(f'{value} out of range for u{self.bits}')

    def serialize(self, value: int) -> int:
        return value

    def deserialize(self, value: int) -> int:
        if type(value) is not int:
            raise ValueError('Value must be an integer')
        return value

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return self.bits


class SignedInteger(ScalePrimitive):
    """
    Signed int type, encoded in little-endian (LE) format
    """

    def __init__(self, bits: int):
        super().__init__()
        self.bits = bits
        self.byte_count = int(self.bits / 8)

    def decode(self, data: ScaleBytes) -> int:
        return int.from_bytes(data.get_next_bytes(self.byte_count), byteorder='little', signed=True)

    def _encode(self, value) -> ScaleBytes:

        if -2**self.bits <= int(value) <= 2**self.bits - 1:
            return ScaleBytes(bytearray(int(value).to_bytes(self.byte_count, 'little', signed=True)))
        else:
            raise ScaleEncodeException(f'{value} out of range for i{self.bits}')

    def serialize(self, value: int) -> int:
        return value

    def deserialize(self, value: int) -> int:
        if type(value) is not int:
            raise ValueError('Value must be an integer')
        return value

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return -self.bits


class Float(ScalePrimitive):

    def __init__(self, bits: int):
        super().__init__()
        self.bits = bits
        self.byte_count = int(self.bits / 8)
        self.struct_format = 'f' if self.bits == 32 else 'd'

    def decode(self, data: ScaleBytes) -> int:
        return struct.unpack(self.struct_format, data.get_next_bytes(self.byte_count))[0]

    def _encode(self, value: float) -> ScaleBytes:
        if type(value) is not float:
            raise ScaleEncodeException(f'{value} is not a float')

        return ScaleBytes(struct.pack(self.struct_format, value))

    def serialize(self, value: float) -> float:
        return value

    def deserialize(self, value: float) -> float:
        if type(value) is not float:
            raise ScaleDeserializeException('Value must be an float')
        return value

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return float(self.bits)


U8 = UnsignedInteger(8)
U16 = UnsignedInteger(16)
U32 = UnsignedInteger(32)
U64 = UnsignedInteger(64)
U128 = UnsignedInteger(128)
U256 = UnsignedInteger(256)

I8 = SignedInteger(8)
I16 = SignedInteger(16)
I32 = SignedInteger(32)
I64 = SignedInteger(64)
I128 = SignedInteger(128)
I256 = SignedInteger(256)
F32 = Float(32)
F64 = Float(64)


class Bool(ScalePrimitive):

    def decode(self, data: ScaleBytes) -> bool:

        bool_data = data.get_next_bytes(1)
        if bool_data not in [b'\x00', b'\x01']:
            raise ScaleDecodeException('Invalid value for datatype "bool"')
        return bool_data == b'\x01'

    def _encode(self, value: bool) -> ScaleBytes:
        if value is True:
            return ScaleBytes('0x01')
        elif value is False:
            return ScaleBytes('0x00')
        else:
            raise ScaleEncodeException("Value must be boolean")

    def serialize(self, value: bool) -> bool:
        return value

    def deserialize(self, value: bool) -> bool:
        return value

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return True


class NullType(ScaleTypeDef):

    def decode(self, data: ScaleBytes) -> any:
        return None

    def _encode(self, value: any) -> ScaleBytes:
        return ScaleBytes(bytearray())

    def serialize(self, value: any) -> any:
        return None

    def deserialize(self, value: any) -> any:
        return None

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return None


Null = NullType()


class StructObject(ScaleType):
    def encode(self, value: Optional[Union[dict, tuple]] = None) -> ScaleBytes:
        if type(value) is tuple:
            # Convert tuple to dict
            try:
                value = {key: value[idx] for idx, key in enumerate(self.type_def.arguments.keys())}
            except IndexError:
                raise ScaleEncodeException("Not enough items in tuple to convert to dict")
        return super().encode(value)


class Struct(ScaleTypeDef):

    arguments = None
    scale_type_cls = StructObject

    def __init__(self, **kwargs):
        if len(kwargs) > 0:
            self.arguments = {key.rstrip('_'): value for key, value in kwargs.items()}
        super().__init__()

    def _encode(self, value: dict) -> ScaleBytes:

        data = ScaleBytes(bytearray())
        for name, scale_obj in self.arguments.items():

            if name not in value:
                raise ScaleEncodeException(f'Argument "{name}" of Struct is missing in given value')

            data += scale_obj.new().encode(value[name])

            if value[name] and issubclass(value[name].__class__, ScaleType):
                value[name] = value[name].serialize()

        return data

    def decode(self, data) -> dict:
        value = {}

        for key, scale_def in self.arguments.items():

            scale_obj = scale_def.new()
            scale_obj.decode(data)

            value[key] = scale_obj

        return value

    def serialize(self, value: dict) -> dict:
        if value is None:
            raise ScaleSerializeException('Value cannot be None')
        return {k: obj.value for k, obj in value.items()}

    def deserialize(self, value: dict) -> dict:
        value_object = {}

        for key, scale_def in self.arguments.items():
            if key in value:
                scale_obj = scale_def.new()

                scale_obj.value_serialized = value[key]
                scale_obj.deserialize(value[key])

                value_object[key] = scale_obj
            else:
                raise ScaleDeserializeException(f'Argument "{key}" of Struct is missing in given value')

        return value_object

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):

        if _recursion_level > max_recursion:
            return f'<{self.__class__.__name__}>'

        return {
            k: scale_def.example_value(_recursion_level + 1, max_recursion) for k, scale_def in self.arguments.items()
        }


class Tuple(ScaleTypeDef):

    values = None

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            self.values = args
        super().__init__()

    def _encode(self, value: tuple) -> ScaleBytes:
        if type(value) is not tuple:
            value = (value,)

        data = ScaleBytes(bytearray())
        for idx, scale_obj in enumerate(self.values):

            data += scale_obj.new().encode(value[idx])
        return data

    def decode(self, data: ScaleBytes) -> tuple:
        value = ()

        for scale_def in self.values:
            scale_obj = scale_def.new()

            scale_obj.decode(data)

            if len(self.values) == 1:
                return scale_obj

            value += (scale_obj,)

        return value

    def serialize(self, value: Union[tuple, ScaleType]) -> tuple:
        if issubclass(value.__class__, ScaleType):
            return value.value

        return tuple((i.value for i in value))

    def deserialize(self, value: tuple) -> tuple:

        if type(value) is not tuple:
            value = (value,)

        value_object = ()

        for idx, scale_def in enumerate(self.values):
            scale_obj = scale_def.new()

            scale_obj.value_serialized = value
            scale_obj.deserialize(value[idx])
            value_object += (scale_obj,)

        return value_object

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return tuple([i.example_value() for i in self.values])

class EnumType(ScaleType):

    @property
    def index(self):
        if self.value_object is not None:
            for index, name in enumerate(self.type_def.variants.keys()):
                if name == self.value_object[0]:
                    return index


class Enum(ScaleTypeDef):

    variants = None

    def __init__(self, **kwargs):
        super().__init__()

        if len(kwargs) > 0:
            self.variants = {key.rstrip('_'): value for key, value in kwargs.items()}

        if self.scale_type_cls is None:
            self.scale_type_cls = EnumType

    def _encode(self, value: Union[str, dict]) -> ScaleBytes:

        # if issubclass(value.__class__, ScaleType) and value.type_def.__class__ is self.__class__:
        #     value = value.value

        if type(value) is dict:
            value = value.copy()

        if type(value) is str:
            # Convert simple enum values
            value = {value: None}

        if type(value) is not dict:
            raise ScaleEncodeException(f"Value must be a dict or str when encoding enums, not '{value}'")

        if len(value) != 1:
            raise ScaleEncodeException("Only one variant can be specified for enums")

        enum_key, enum_value = list(value.items())[0]

        for idx, (variant_name, variant_obj) in enumerate(self.variants.items()):

            if enum_key == variant_name:

                data = ScaleBytes(bytearray([idx]))

                if variant_obj is not None:

                    data += variant_obj.new().encode(enum_value)

                return data

        raise ScaleEncodeException(f"Variant '{enum_key}' not defined for this enum")

    def decode(self, data: ScaleBytes) -> tuple:

        index = int.from_bytes(data.get_next_bytes(1), byteorder='little')

        try:
            enum_key, enum_variant = list(self.variants.items())[index]
        except IndexError:
            raise ScaleDecodeException(f"Index '{index}' not present in Enum type mapping")

        if enum_variant is None:
            return (enum_key, None)
        else:
            scale_obj = enum_variant.new()
            scale_obj.decode(data)
            return (enum_key, scale_obj)

    def serialize(self, value: tuple) -> Union[str, dict]:
        if value[1] is None:
            return value[0]
        else:
            return {value[0]: value[1].value}

    def deserialize(self, value: Union[str, dict]) -> tuple:
        if type(value) is str:
            value = {value: None}

        if isinstance(value, enum.Enum):
            value = {value.name: None}

        if len(list(value.items())) != 1:
            raise ScaleDeserializeException("Only one variant can be specified for enums")

        enum_key, enum_value = list(value.items())[0]

        for idx, (variant_name, variant_obj) in enumerate(self.variants.items()):

            if enum_key == variant_name:

                if variant_obj is not None:
                    enum_value_obj = variant_obj.new()
                    enum_value_obj.value_serialized = enum_value
                    enum_value_obj.deserialize(enum_value)
                else:
                    enum_value_obj = None

                return (enum_key, enum_value_obj)

        raise ValueError(f"Error while deserializing Enum; variant '{enum_key}' not found")

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):

        if _recursion_level > max_recursion:
            return f'<{self.__class__.__name__}>'

        example = {}
        for idx, (variant_name, variant_obj) in enumerate(self.variants.items()):
            if not variant_name.startswith('__'):
                if variant_obj is None:
                    example[variant_name] = None
                else:
                    example[variant_name] = variant_obj.example_value(_recursion_level + 1, max_recursion)
        return example


class Option(ScaleTypeDef):
    def __init__(self, some):
        self.some = some
        super().__init__()

    def _encode(self, value: any) -> ScaleBytes:
        if value is None:
            return ScaleBytes('0x00')
        else:
            return ScaleBytes('0x01') + self.some.encode(value, external_call=False)

    def decode(self, data: ScaleBytes) -> Optional[ScaleType]:
        if data.get_next_bytes(1) == b'\x00':
            return None
        else:
            scale_obj = self.some.new()
            scale_obj.decode(data)
            return scale_obj

    def serialize(self, value: Optional[ScaleType]) -> any:
        if value is not None:
            return value.value

    def deserialize(self, value: any) -> Optional[ScaleType]:
        if value is not None:
            some_obj = self.some.new()
            some_obj.deserialize(value)
            return some_obj

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return None, self.some.example_value()


class Compact(ScaleTypeDef):
    def __init__(self, type_: ScaleTypeDef = None):
        self.type = type_
        self.compact_length = 0
        self.compact_bytes = None
        super().__init__()

    def process_compact_bytes(self, data):
        compact_byte = data.get_next_bytes(1)
        try:
            byte_mod = compact_byte[0] % 4
        except IndexError:
            raise ScaleDecodeException("Invalid byte for Compact")

        if byte_mod == 0:
            self.compact_length = 1
        elif byte_mod == 1:
            self.compact_length = 2
        elif byte_mod == 2:
            self.compact_length = 4
        else:
            self.compact_length = int(5 + (compact_byte[0] - 3) / 4)

        if self.compact_length == 1:
            self.compact_bytes = compact_byte
        elif self.compact_length in [2, 4]:
            self.compact_bytes = compact_byte + data.get_next_bytes(self.compact_length - 1)
        else:
            self.compact_bytes = data.get_next_bytes(self.compact_length - 1)

        return self.compact_bytes

    def decode(self, data: ScaleBytes) -> any:
        self.process_compact_bytes(data)

        if self.compact_length <= 4:
            return int(int.from_bytes(self.compact_bytes, byteorder='little') / 4)
        else:
            return int.from_bytes(self.compact_bytes, byteorder='little')

    def _encode(self, value: int) -> ScaleBytes:

        value = int(value)

        if value <= 0b00111111:
            return ScaleBytes(bytearray(int(value << 2).to_bytes(1, 'little')))

        elif value <= 0b0011111111111111:
            return ScaleBytes(bytearray(int((value << 2) | 0b01).to_bytes(2, 'little')))

        elif value <= 0b00111111111111111111111111111111:
            return ScaleBytes(bytearray(int((value << 2) | 0b10).to_bytes(4, 'little')))

        else:
            for bytes_length in range(4, 68):
                if 2 ** (8 * (bytes_length - 1)) <= value < 2 ** (8 * bytes_length):
                    return ScaleBytes(bytearray(
                        ((bytes_length - 4) << 2 | 0b11).to_bytes(1, 'little') + value.to_bytes(bytes_length,
                                                                                                'little')))
            else:
                raise ScaleEncodeException('{} out of range'.format(value))

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return 1

    def serialize(self, value: int) -> int:
        return value

    def deserialize(self, value: int) -> int:
        if type(value) is not int:
            raise ValueError('Value must be an integer')
        return value


class VecType(ScaleType):

    def __len__(self):
        return len(self.value_object)


class Vec(ScaleTypeDef):
    def __init__(self, type_def: ScaleTypeDef):
        super().__init__()
        self.scale_type_cls = VecType
        self.type_def = type_def

    def _encode(self, value: list) -> ScaleBytes:

        if self.type_def is U8:
            return Bytes.encode(value, external_call=False)

        # Encode length of Vec
        data = Compact().new().encode(len(value))

        for idx, item in enumerate(value):
            data += self.type_def.new().encode(item)
            if item and issubclass(item.__class__, ScaleType):
                value[idx] = item.serialize()

        return data

    def decode(self, data: ScaleBytes) -> list:

        if self.type_def is U8:
            return Bytes.decode(data)

        # Decode length of Vec
        length = Compact().decode(data)

        value = []

        for _ in range(0, length):
            obj = self.type_def.new()
            obj.decode(data)

            value.append(obj)

        return value

    def serialize(self, value: list) -> list:
        if self.type_def is U8:
            return Bytes.serialize(value)
        return [i.value_serialized for i in value]

    def deserialize(self, value: list) -> list:
        if self.type_def is U8:
            return Bytes.deserialize(value)


        value_object = []

        for item in value:
            obj = self.type_def.new()
            obj.value_serialized = item
            obj.deserialize(item)

            value_object.append(obj)

        return value_object

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        if self.type_def is U8:
            return b'Bytes'
        return [self.type_def.example_value()]


class BitVec(ScaleTypeDef):
    """
    A BitVec that represents an array of bits. The bits are however stored encoded. The difference between this
    and a normal Bytes would be that the length prefix indicates the number of bits encoded, not the bytes
    """

    def _encode(self, value: Union[list, str, int]) -> ScaleBytes:

        if type(value) is list:
            value = sum(v << i for i, v in enumerate(reversed(value)))

        if type(value) is str and value[0:2] == '0b':
            value = int(value[2:], 2)

        if type(value) is not int:
            raise ScaleEncodeException("Provided value is not an int, binary str or a list of booleans")

        if value == 0:
            return ScaleBytes(b'\x00')

        # encode the length in a compact u32
        data = Compact().encode(value.bit_length(), external_call=False)

        byte_length = math.ceil(value.bit_length() / 8)

        return data + value.to_bytes(length=byte_length, byteorder='little')

    def decode(self, data: ScaleBytes) -> str:
        # Decode length of Vec
        length = Compact().decode(data)

        total = math.ceil(length / 8)

        value_int = int.from_bytes(data.get_next_bytes(total), byteorder='little')

        return '0b' + bin(value_int)[2:].zfill(length)

    def serialize(self, value: str) -> str:
        return value

    def deserialize(self, value: str) -> str:
        return value


class ArrayObject(ScaleType):

    def to_bytes(self) -> bytes:
        if self.type_def.type_def is not U8:
            raise ScaleDeserializeException('Only an Array of U8 can be represented as bytes')
        return self.value_object


class Array(ScaleTypeDef):

    scale_type_cls = ArrayObject

    def __init__(self, type_def: ScaleTypeDef, length: int):
        self.type_def = type_def
        self.length = length
        super().__init__()

    def _encode(self, value: Union[list, str, bytes]) -> ScaleBytes:

        if self.type_def is U8:

            if type(value) is list:
                value = bytes(value)
            elif type(value) is str:
                if value[0:2] == '0x':
                    value = bytes.fromhex(value[2:])
                else:
                    value = value.encode('utf-8')

            if type(value) is not bytes:
                raise ScaleEncodeException('value should be of type list, str or bytes')

            if len(value) != self.length:
                raise ScaleEncodeException(f'Value should be {self.length} bytes long')

            return ScaleBytes(value)
        else:
            data = ScaleBytes(bytearray())

            if type(value) is not list:
                raise ScaleEncodeException("Value must be of type list")

            if len(value) != self.length:
                raise ScaleEncodeException("Length of list does not match size of array")

            for item in value:
                data += self.type_def.encode(item, external_call=False)

            return data

    def decode(self, data: ScaleBytes) -> Union[list, bytes]:
        if self.type_def is U8:
            return data.get_next_bytes(self.length)
        else:
            value = []

            for _ in range(0, self.length):
                obj = self.type_def.new()
                obj.decode(data)

                value.append(obj)

            return value

    def serialize(self, value: Union[list, bytes]) -> Union[list, str]:
        if type(value) is list:
            return [i.value_serialized for i in value]
        else:
            return f'0x{value.hex()}'

    def deserialize(self, value: Union[list, str, bytes, bytearray]) -> Union[list, bytes]:

        if type(value) not in [list, str, bytes, bytearray]:
            raise ScaleDeserializeException('value should be of type list, str or bytes')

        if type(value) is str:
            if value[0:2] == '0x':
                value = bytes.fromhex(value[2:])
            else:
                value = value.encode()

        if len(value) != self.length:
            raise ScaleDeserializeException('Length of array does not match size of value')

        if type(value) is bytearray:
            value = bytes(value)

        if type(value) is bytes:
            if self.type_def is not U8:
                raise ScaleDeserializeException('Only an Array of U8 can be represented as (hex)bytes')

            return value

        if type(value) is list:

            value_object = []

            for item in value:
                obj = self.type_def.new()
                obj.value_serialized = item
                obj.deserialize(item)

                value_object.append(obj)

            return value_object

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        if self.type_def is U8:
            return f'0x{str(self.length).zfill(2) * self.length}'
        else:
            return [self.type_def.example_value()] * self.length


class DataType(EnumType):

    def decode(self, data: ScaleBytes, check_remaining=False) -> any:
        value = super().decode(data)

        if type(value) is dict:
            items = list(value.items())[0]
            if items[0].startswith('Raw'):
                self.value_serialized = {'Raw': items[1]}
                self.value_object = ('Raw', self.value_object[1])

        return self.value_serialized

    def encode(self, value: dict) -> ScaleBytes:
        items = list(value.items())[0]
        if items[0] == 'Raw':
            value = {f'Raw{len(items[1])}': items[1]}

        return super().encode(value)


class Map(ScaleTypeDef):
    def __init__(self, key_def: ScaleTypeDef, value_def: ScaleTypeDef):
        super().__init__()
        # self.scale_type_cls = VecType
        self.key_def = key_def
        self.value_def = value_def

    def _encode(self, value: list) -> ScaleBytes:
        # Encode length of Vec
        data = Compact().encode(len(value), external_call=False)

        for item_key, item_value in value:
            data += self.key_def.encode(item_key, external_call=False)
            data += self.value_def.encode(item_value, external_call=False)

        return data

    def decode(self, data: ScaleBytes) -> list:
        # Decode length of Map
        length = Compact().decode(data)

        value = []

        for _ in range(0, length):
            key_obj = self.key_def.new()
            key_obj.decode(data)
            value_obj = self.value_def.new()
            value_obj.decode(data)
            value.append((key_obj, value_obj))

        return value

    def serialize(self, value: list) -> list:
        output = []
        for k, v in value:
            if type(k) is ScaleType and type(v) is ScaleType:
                output.append((k.value_serialized, v.value_serialized))
            else:
                output.append((k, v))
        return output

    def deserialize(self, value: list) -> list:
        return [(self.key_def.deserialize(k), self.value_def.deserialize(v)) for k, v in value]


class BytesDef(ScaleTypeDef):
    """
    A variable collection of bytes, stored as an `Vec<u8>`
    """

    def _encode(self, value: Union[str, bytes, bytearray, list]) -> ScaleBytes:

        if type(value) is str:
            if value[0:2] == '0x':
                # TODO implicit HexBytes conversion can have unexpected result if string is actually starting with '0x'
                value = bytes.fromhex(value[2:])
            else:
                value = value.encode('utf-8')

        elif type(value) in (bytearray, list):
            value = bytes(value)

        if type(value) is not bytes:
            raise ScaleEncodeException(f'Cannot encode type "{type(value)}"')

        # Encode length of Vec
        data = Compact().new().encode(len(value))

        return data + value

    def decode(self, data: ScaleBytes) -> bytearray:
        # Decode length of Vec
        length = Compact().decode(data)

        return data.get_next_bytes(length)

    def serialize(self, value: bytearray) -> str:
        return f'0x{value.hex()}'

    def deserialize(self, value: Union[bytes, str, list]) -> bytes:

        if type(value) in (list, bytearray):
            value = bytes(value)

        elif type(value) is str:
            if value[0:2] == '0x':
                value = bytes.fromhex(value[2:])
            else:
                value = value.encode('utf-8')

        if type(value) is not bytes:
            raise ScaleDeserializeException(f'Cannot deserialize type "{type(value)}"')

        return value

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return b'Bytes'


class StringDef(BytesDef):
    def decode(self, data: ScaleBytes) -> str:
        value = super().decode(data)

        try:
            return value.decode()
        except UnicodeDecodeError:
            return '0x{}'.format(value.hex())

    def serialize(self, value: str) -> str:
        return value

    def deserialize(self, value: str) -> str:
        return value

    def create_example(self, _recursion_level: int = 0):
        return 'String'


class HashDefObject(ScaleType):
    def to_bytes(self) -> bytes:
        return self.value_object


class HashDef(ScaleTypeDef):

    scale_type_cls = HashDefObject

    def __init__(self, bits: int):
        super().__init__()
        self.bits = bits
        self.byte_count = int(self.bits / 8)

    def decode(self, data: ScaleBytes) -> bytes:
        return data.get_next_bytes(self.byte_count)

    def _encode(self, value: Union[str, bytes]) -> ScaleBytes:

        if type(value) is str:
            if value[0:2] != '0x' or len(value) != (self.byte_count*2)+2:
                raise ScaleEncodeException(f'Value should start with "0x" and should be {self.byte_count} bytes long')

            value = bytes.fromhex(value[2:])

        if type(value) is not bytes:
            raise ScaleEncodeException('value should be of type str or bytes')

        if len(value) != self.byte_count:
            raise ScaleEncodeException(f'Value should be {self.byte_count} bytes long')

        return ScaleBytes(value)

    def serialize(self, value: bytes) -> str:
        return f'0x{value.hex()}'

    def deserialize(self, value: Union[str, bytes, bytearray]) -> bytes:
        if type(value) is str:
            value = bytes.fromhex(value[2:])

        if type(value) is bytearray:
            value = bytes(value)

        if type(value) is not bytes:
            raise ScaleDeserializeException('value should be of type str or bytes')

        return value

    def example_value(self, _recursion_level: int = 0, max_recursion: int = TYPE_DECOMP_MAX_RECURSIVE):
        return f'0x{str(self.byte_count).zfill(2) * self.byte_count}'


String = StringDef()
Bytes = BytesDef()
Type = String
Text = String
H256 = HashDef(256)
H512 = HashDef(512)
Hash = H256
HashMap = Map
BTreeMap = Map



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.types import Enum, Bool, U32


class TestEnum(unittest.TestCase):

    def test_enum_encode_decode(self):
        scale_obj = Enum(Bool=Bool(), Number=U32, None_=None).new()
        value = {'Bool': True}

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

        value = {'Number': 7643}

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

        value = 'None'

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

    def test_enum_deserialize(self):
        scale_obj = Enum(Bool=Bool(), Number=U32, None_=None).new()

        scale_obj.deserialize({'Bool': True})
        self.assertEqual(('Bool', Bool().new(value=True)), scale_obj.value_object)

        scale_obj.deserialize({'Number': 1})
        self.assertEqual(('Number', U32.new(value=1)), scale_obj.value_object)

        scale_obj.deserialize({'None': None})
        self.assertEqual(('None', None), scale_obj.value_object)

        scale_obj.deserialize('None')
        self.assertEqual(('None', None), scale_obj.value_object)


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import U16, I16, U8


class TestInteger(unittest.TestCase):

    def test_u16_decode(self):
        obj = U16.new()
        obj.decode(ScaleBytes("0x2efb"))
        self.assertEqual(obj.value, 64302)

    def test_i16_decode(self):
        obj = I16.new()
        obj.decode(ScaleBytes("0x2efb"))
        self.assertEqual(obj.value, -1234)

    def test_u16(self):
        obj = U16.new()
        obj.encode(64302)
        self.assertEqual(str(obj.data), "0x2efb")

    def test_i16_encode(self):
        obj = I16.new()
        obj.encode(-1234)
        self.assertEqual(str(obj.data), "0x2efb")

    def test_i16_encode_out_of_bounds(self):
        obj = I16.new()
        self.assertRaises(OverflowError, obj.encode, -32769)

    def test_u8(self):
        scale_obj = U8.new()
        value = 42

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.types import Option, U16, Bytes, String


class TestOption(unittest.TestCase):

    def test_option(self):
        scale_obj = Option(U16).new()

        value = None

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

        value = 12788

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

    def test_option_empty_encode_decode(self):

        value = None

        obj = Option(Bytes).new()
        data = obj.encode(value)

        self.assertEqual(obj.decode(data), value)

    def test_option_string_encode_decode(self):
        value = "Test"

        obj = Option(String).new()
        data = obj.encode(value)

        self.assertEqual(obj.decode(data), value)


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import Vec, String, U16


class TestVec(unittest.TestCase):

    def test_vec_string_encode_decode(self):

        value = ['test', 'vec']

        obj = Vec(String).new()
        data = obj.encode(value)

        obj = Vec(String).new()

        self.assertEqual(value, obj.decode(data))

    def test_vec_integer(self):

        obj = Vec(U16).new()

        value = [1, 2]
        data = obj.encode(value)

        self.assertEqual(data, ScaleBytes('0x0801000200'))
        self.assertEqual(value, obj.decode(data))

        self.assertEqual(obj.value_object[0].value, 1)
        self.assertEqual(obj.value_object[1].value, 2)


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.exceptions import ScaleDecodeException
from scalecodec.types import Bool


class TestBoolean(unittest.TestCase):

    def test_bool_true(self):
        obj = Bool().new()
        obj.decode(ScaleBytes("0x01"))
        self.assertEqual(obj.value, True)

    def test_bool_false(self):
        obj = Bool().new()
        obj.decode(ScaleBytes("0x00"))
        self.assertEqual(obj.value, False)

    def test_bool_invalid(self):
        obj = Bool().new()
        self.assertRaises(ScaleDecodeException, obj.decode, ScaleBytes("0x02"))

    def test_bool_encode_decode(self):
        scale_obj = Bool().new()
        value = True

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

    def test_bool_encode_false(self):
        scale_obj = Bool().new()
        data = scale_obj.encode(False)
        self.assertEqual(ScaleBytes("0x00"), data)


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import String, Bytes


class TestBytes(unittest.TestCase):

    def test_bytes_encode_decode(self):
        scale_obj = Bytes.new()
        value = "0x1274657374"

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

    def test_bytes_encode_bytes(self):
        value = b'This is a test'

        obj = Bytes.new()
        data = obj.encode(value)

        self.assertEqual("0x385468697320697320612074657374", data.to_hex())

    def test_bytes_encode_bytearray(self):
        value = bytearray(b'This is a test')

        obj = Bytes.new()
        data = obj.encode(value)

        self.assertEqual("0x385468697320697320612074657374", data.to_hex())

    def test_bytes_encode_list_of_u8(self):
        value = [84, 104, 105, 115, 32, 105, 115, 32, 97, 32, 116, 101, 115, 116]

        obj = Bytes.new()
        data = obj.encode(value)

        self.assertEqual("0x385468697320697320612074657374", data.to_hex())

    def test_hexbytes_encode_decode(self):

        value = '0x5468697320697320612074657374'

        obj = Bytes.new()
        data = obj.encode(value)

        obj_check = Bytes.new()

        self.assertEqual(obj_check.decode(data), value)

if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import String


class TestString(unittest.TestCase):

    def test_string_decode_encode(self):
        obj = String.new()
        obj.decode(ScaleBytes("0x1054657374"))
        self.assertEqual(obj.value, "Test")

        data = obj.encode("Test")

        self.assertEqual("0x1054657374", data.to_hex())

    def test_string_encode_decode(self):

        value = 'This is a test'

        obj = String.new()
        data = obj.encode(value)

        obj_check = String.new()

        self.assertEqual(obj_check.decode(data), value)

    def test_string_multibyte_chars(self):
        obj = String.new()

        data = obj.encode('µ')
        self.assertEqual('0x08c2b5', data.to_hex())

        obj.decode(data)
        self.assertEqual(obj.value, "µ")


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2020 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#  test_scalebytes.py
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.exceptions import RemainingScaleBytesNotEmptyException
from scalecodec.types import Compact, U32, Array, U8, String


class TestScaleBytes(unittest.TestCase):

    def test_unknown_data_format(self):
        self.assertRaises(ValueError, ScaleBytes, 123)
        self.assertRaises(ValueError, ScaleBytes, 'test')

    def test_bytes_data_format(self):
        obj = Compact(U32).new()
        obj.decode(ScaleBytes(b"\x02\x09\x3d\x00"))
        self.assertEqual(obj.value, 1000000)

    def test_remaining_bytes(self):
        scale = ScaleBytes("0x01020304")
        scale.get_next_bytes(1)
        self.assertEqual(scale.get_remaining_bytes(), b'\x02\x03\x04')

    def test_reset(self):
        scale = ScaleBytes("0x01020304")
        scale.get_next_bytes(1)
        scale.reset()
        self.assertEqual(scale.get_remaining_bytes(), b'\x01\x02\x03\x04')

    def test_add_scalebytes(self):
        scale_total = ScaleBytes("0x0102") + "0x0304"

        self.assertEqual(scale_total.data, bytearray.fromhex("01020304"))

    def test_scale_bytes_compare(self):
        self.assertEqual(ScaleBytes('0x1234'), ScaleBytes('0x1234'))
        self.assertNotEqual(ScaleBytes('0x1234'), ScaleBytes('0x555555'))

    def test_no_more_bytes_available(self):
        obj = Array(U8, 4).new()
        with self.assertRaises(RemainingScaleBytesNotEmptyException):
            obj.decode(ScaleBytes("0x010203"), check_remaining=False)




# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import Tuple, Compact, U32, U8


class TestTuple(unittest.TestCase):

    def test_tuple_decode(self):
        obj = Tuple(Compact(U32), Compact(U32)).new()
        obj.decode(ScaleBytes("0x0c00"))
        self.assertEqual(obj.value, (3, 0))

    def test_tuple_encode_decode(self):
        scale_obj = Tuple(U8, U8).new()
        value = (1, 5)

        data = scale_obj.encode(value)
        scale_obj.decode(data)

        self.assertEqual(value, scale_obj.value)

    def test_tuple_deserialize(self):
        obj = Tuple(Compact(U32), Compact(U32)).new()
        obj.deserialize((3, 2))
        self.assertEqual(obj.value, (3, 2))

    def test_tuple_single_value(self):
        # PolkadotJS compatilibity
        obj = Tuple(U8).new()
        obj.decode(ScaleBytes('0x03'))
        self.assertEqual(obj.value, 3)
        self.assertEqual(obj.value_object, U8.new(value=3))


if __name__ == '__main__':
    unittest.main()



import dataclasses
from dataclasses import is_dataclass
import enum
from typing import Type, TypeVar, Union
import typing
import json

from scalecodec.base import ScaleTypeDef, ScaleType, ScaleBytes
from scalecodec.types import Struct, Option, Vec, Enum

T = TypeVar('T')


class ScaleSerializable:
    @classmethod
    def scale_type_def(cls) -> ScaleTypeDef:
        if is_dataclass(cls):

            arguments = {}
            for field in dataclasses.fields(cls):
                arguments[field.name] = cls.dataclass_field_to_scale_typ_def(field)

            return Struct(**arguments)
        elif issubclass(cls, enum.Enum):
            variants = {status.name: None for status in cls}
            return Enum(**variants)

        raise NotImplementedError

    def serialize(self) -> Union[str, int, float, bool, dict, list]:
        scale_type = self.to_scale_type()
        return scale_type.serialize()

    @classmethod
    def deserialize(cls: Type[T], data: Union[str, int, float, bool, dict, list]) -> T:
        scale_type = cls.scale_type_def().new()
        scale_type.deserialize(data)
        return cls.from_scale_type(scale_type)

    def to_scale_type(self) -> ScaleType:

        if not is_dataclass(self) and not issubclass(self.__class__, enum.Enum):
            raise NotImplementedError("Type not supported.")

        scale_type = self.scale_type_def().new()

        if issubclass(self.__class__, enum.Enum):
            scale_type.deserialize(self.name)
        elif is_dataclass(self):
            value = {}
            for field in dataclasses.fields(self):

                actual_type = field.type
                field_name = field.name[:-1] if field.name.endswith('_') else field.name

                if typing.get_origin(actual_type) is typing.Union:
                    # Extract the arguments of the Union type
                    args = typing.get_args(actual_type)
                    if type(None) in args:
                        # If NoneType is in the args, it's an Optional
                        actual_type = [arg for arg in args if arg is not type(None)][0]

                if getattr(self, field.name) is None:
                    value[field_name] = None
                else:

                    if typing.get_origin(actual_type) is list:
                        actual_type = typing.get_args(actual_type)[0]

                        if issubclass(actual_type, ScaleSerializable):
                            value[field_name] = [i.serialize() for i in getattr(self, field.name)]
                        else:
                            value[field_name] = getattr(self, field.name)

                    # TODO too simplified now
                    elif issubclass(actual_type, ScaleSerializable):

                        value[field_name] = getattr(self, field.name).serialize()
                    else:
                        value[field_name] = getattr(self, field.name)

            scale_type.deserialize(value)

        return scale_type

    @classmethod
    def from_scale_type(cls: Type[T], scale_type: ScaleType) -> T:
        if is_dataclass(cls):

            fields = {}

            for field in dataclasses.fields(cls):

                scale_field_name = field.name[:-1] if field.name.endswith('_') else field.name

                actual_type = field.type

                if typing.get_origin(field.type) is typing.Union:
                    # Extract the arguments of the Union type
                    args = typing.get_args(field.type)
                    if type(None) in args:
                        # If NoneType is in the args, it's an Optional
                        if field.name in scale_type.value:
                            if scale_type.value[field.name] is None:
                                fields[field.name] = None
                                continue
                            else:
                                actual_type = [arg for arg in args if arg is not type(None)][0]
                        else:
                            # print(field.name)
                            continue

                if typing.get_origin(actual_type) is list:
                    items = []
                    actual_type = typing.get_args(actual_type)[0]

                    if issubclass(type(scale_type.type_def), (Struct, Option)):
                        list_items = scale_type.value_object[scale_field_name].value_object
                    elif issubclass(type(scale_type.type_def), (Vec, Enum)):
                        list_items = scale_type.value_object[1].value_object
                    else:
                        raise ValueError(f'Unsupported type: {type(scale_type.type_def)}')

                    for item in list_items:
                        if actual_type in [str, int, float, bool]:
                            items.append(item.value)
                        elif actual_type is bytes:
                            items.append(item.to_bytes())
                        elif is_dataclass(actual_type):
                            items.append(actual_type.from_scale_type(item))

                    fields[field.name] = items

                elif actual_type in [str, int, float, bool]:
                    fields[field.name] = scale_type.value[scale_field_name]
                elif actual_type is bytes:
                    fields[field.name] = scale_type.value_object[scale_field_name].to_bytes()
                elif is_dataclass(actual_type):
                    try:

                        # TODO unwrap Option
                        if issubclass(type(scale_type.type_def), (Struct, Option)):

                            field_scale_type = scale_type.value_object[scale_field_name]
                        elif issubclass(type(scale_type.type_def), Enum):
                            field_scale_type = scale_type.value_object[1]
                        else:
                            raise ValueError(f"Unexpected type {type(scale_type.type_def)}")

                        fields[field.name] = actual_type.from_scale_type(field_scale_type)
                    except (KeyError, TypeError) as e:
                        print('oeps', str(e))
                elif issubclass(actual_type, enum.Enum):
                    fields[field.name] = actual_type[scale_type.value_object[1].value]
            return cls(**fields)
        raise NotImplementedError

    def to_scale_bytes(self) -> ScaleBytes:
        scale_obj = self.to_scale_type()
        return scale_obj.encode()

    @classmethod
    def from_scale_bytes(cls: Type[T], scale_bytes: ScaleBytes) -> T:
        scale_obj = cls.scale_type_def().new()
        scale_obj.decode(scale_bytes)
        return cls.from_scale_type(scale_obj)

    def to_json(self) -> str:
        return json.dumps(self.serialize(), indent=4)

    @classmethod
    def from_json(cls: Type[T], json_data: str) -> T:
        # data = json.loads(json_data)
        return cls.deserialize(json_data)

    @classmethod
    def dataclass_field_to_scale_typ_def(cls, field) -> ScaleTypeDef:

        if 'scale' in field.metadata:
            return field.metadata['scale']

        # Check if the field type is an instance of Optional
        actual_type = field.type
        wrap_option = False
        wrap_vec = False

        if typing.get_origin(field.type) is typing.Union:
            # Extract the arguments of the Union type
            args = typing.get_args(field.type)
            if type(None) in args:
                # If NoneType is in the args, it's an Optional
                wrap_option = True
                actual_type = [arg for arg in args if arg is not type(None)][0]
                # print(f"The field '{field.name}' is Optional with inner type: {actual_type}")

        if typing.get_origin(actual_type) is list:
            wrap_vec = True
            actual_type = typing.get_args(actual_type)[0]

        if is_dataclass(actual_type):
            if issubclass(actual_type, ScaleSerializable):
                scale_def = actual_type.scale_type_def()
            else:
                raise ValueError(f"Cannot serialize dataclass {field.type.__class__}")

        elif actual_type is bytes:
            raise ValueError("bytes is ambiguous; specify SCALE type def in metadata e.g. {'scale': H256}")
        elif actual_type is int:
            raise ValueError("int is ambiguous; specify SCALE type def in metadata e.g. {'scale': U32}")

        elif issubclass(actual_type, enum.Enum):
            variants = {status.name: None for status in actual_type}
            scale_def = Enum(**variants)

        else:
            raise ValueError(f"Cannot convert {actual_type} to ScaleTypeDef")

        if wrap_vec:
            scale_def = Vec(scale_def)
        if wrap_option:
            scale_def = Option(scale_def)

        return scale_def



import enum
import json
import os
import unittest
from dataclasses import dataclass, field
from os import path
from typing import Optional, Type, Union, List

from scalecodec.base import ScaleBytes, ScaleType
from scalecodec.mixins import ScaleSerializable, T
from scalecodec.types import H256, U8, Array, Enum


# Test definitions


@dataclass
class ValidatorData(ScaleSerializable):
    bandersnatch: bytes = field(metadata={'scale': H256})
    ed25519: bytes = field(metadata={'scale': H256})
    bls: bytes = field(metadata={'scale': Array(U8, 144)})
    metadata: bytes = field(metadata={'scale': Array(U8, 128)})


@dataclass
class EpochMark(ScaleSerializable):
    entropy: bytes = field(metadata={'scale': H256})
    validators: List[bytes] = field(metadata={'scale': Array(H256, 6)})

@dataclass
class OutputMarks(ScaleSerializable):
    epoch_mark: Optional[EpochMark] = None


class CustomErrorCode(ScaleSerializable, enum.Enum):
    bad_slot = 0  # Timeslot value must be strictly monotonic.
    unexpected_ticket = 1  # Received a ticket while in epoch's tail.
    bad_ticket_order = 2  # Tickets must be sorted.
    bad_ticket_proof = 3  # Invalid ticket ring proof.
    bad_ticket_attempt = 4  # Invalid ticket attempt value.
    reserved = 5  # Reserved
    duplicate_ticket = 6  # Found a ticket duplicate.
    too_many_tickets = 7  # Found amount of tickets > K


@dataclass
class Output(ScaleSerializable):
    ok: Optional[OutputMarks] = None  # Markers
    err: Optional[CustomErrorCode] = None

    @classmethod
    def scale_type_def(cls):

        return Enum(
            ok=OutputMarks.scale_type_def(),
            err=CustomErrorCode.scale_type_def()
        )

    def to_scale_type(self) -> ScaleType:
        scale_type = self.scale_type_def().new()
        scale_type.deserialize(self.serialize())
        return scale_type

    @classmethod
    def deserialize(cls: Type[T], data: Union[str, int, float, bool, dict, list]) -> T:

        return super().deserialize(data)

    def serialize(self) -> Union[str, int, float, bool, dict, list]:
        if self.err is not None:
            return {'err': self.err.serialize()}
        else:
            return {'ok': self.ok.serialize()}


class TestSerializableMixin(unittest.TestCase):

    def setUp(self):
        data = {
            'bandersnatch': '0x5e465beb01dbafe160ce8216047f2155dd0569f058afd52dcea601025a8d161d',
            'ed25519': '0x3b6a27bcceb6a42d62a3a8d02a6f0d73653215771de243a63ac048a18b59da29',
            'bls': '0x000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000',
            'metadata': '0x0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'
        }

        self.test_obj = ValidatorData.deserialize(data)

    def test_dataclass_serialization(self):
        output = Output(ok=OutputMarks(epoch_mark=None))
        value = output.serialize()
        self.assertEqual({'ok': {'epoch_mark': None}}, value)

        output = Output(err=CustomErrorCode.duplicate_ticket)
        value = output.serialize()

        self.assertEqual({'err': 'duplicate_ticket'}, value)

    def test_dataclass_to_scale_type(self):
        output = Output(
            ok=OutputMarks(
                epoch_mark=EpochMark(
                    entropy=bytes(32),
                    validators=[bytes(32), bytes(32), bytes(32), bytes(32), bytes(32), bytes(32)]
                )
            )
        )
        scale_type = output.to_scale_type()
        output2 = Output.from_scale_type(scale_type)
        self.assertEqual(output, output2)

    def test_deserialize(self):

        data = {
            'bandersnatch': '0x5e465beb01dbafe160ce8216047f2155dd0569f058afd52dcea601025a8d161d',
            'ed25519': '0x3b6a27bcceb6a42d62a3a8d02a6f0d73653215771de243a63ac048a18b59da29',
            'bls': '0x000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000',
            'metadata': '0x0000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'
        }

        validator_obj = ValidatorData.deserialize(data)

        self.assertEqual(self.test_obj, validator_obj)
        self.assertEqual(data, validator_obj.serialize())

    def test_from_to_scale_bytes(self):

        scale_data = self.test_obj.to_scale_bytes()

        validator_obj = ValidatorData.from_scale_bytes(scale_data)

        self.assertEqual(self.test_obj, validator_obj)


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import Compact, U32, U16, Tuple


class TestScaleTypes(unittest.TestCase):

    def test_multiple_decode_without_error(self):
        obj = U16.new()
        obj.decode(ScaleBytes("0x2efb"))
        obj.decode(ScaleBytes("0x2efb"))
        self.assertEqual(obj.value, 64302)

    def test_value_equals_value_serialized_and_value_object(self):
        obj = Tuple(Compact(U32), Compact(U32)).new()
        obj.decode(ScaleBytes("0x0c00"))
        self.assertEqual(obj.value, obj.value_serialized)
        self.assertEqual(obj.value, obj.value_object)

    def test_value_object(self):
        obj = Tuple(Compact(U32), Compact(U32)).new()
        obj.decode(ScaleBytes("0x0c00"))
        self.assertEqual(obj.value_object[0].value_object, 3)
        self.assertEqual(obj.value_object[1].value_object, 0)

    def test_value_object_shorthand(self):
        obj = Tuple(Compact(U32), Compact(U32)).new()
        obj.decode(ScaleBytes("0x0c00"))
        self.assertEqual(obj[0], 3)
        self.assertEqual(obj[1], 0)



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.exceptions import ScaleDecodeException, RemainingScaleBytesNotEmptyException
from scalecodec.types import Compact, U32, U128


class TestCompact(unittest.TestCase):

    def test_compact_u32(self):
        obj = Compact(U32).new()
        obj.decode(ScaleBytes("0x02093d00"))
        self.assertEqual(obj.value, 1000000)

    def test_compact_u32_1byte(self):
        obj = Compact(U32).new()
        obj.decode(ScaleBytes("0x18"))
        self.assertEqual(obj.value, 6)

    def test_compact_u32_remaining_bytes(self):
        obj = Compact(U32).new()
        with self.assertRaises(ScaleDecodeException):
            obj.decode(ScaleBytes("0x02093d0001"), check_remaining=True)

    def test_compact_u32_invalid(self):
        obj = Compact(U32).new()
        self.assertRaises(RemainingScaleBytesNotEmptyException, obj.decode, ScaleBytes("0x"))

    def test_compact_u32_1byte_encode(self):
        obj = Compact(U32).new()
        obj.decode(ScaleBytes("0x18"))

        obj = Compact(U32).new()
        obj.encode(6)
        self.assertEqual(str(obj.data), "0x18")

    def test_compact_u32_2bytes_encode(self):
        obj = Compact(U32).new()
        obj.encode(6000)
        self.assertEqual(str(obj.data), "0xc15d")

    def test_compact_u32_4bytes_encode(self):

        obj = Compact(U32).new()
        obj.encode(1000000)
        self.assertEqual(str(obj.data), "0x02093d00")

    def test_compact_u32_larger_than_4bytes_encode(self):

        obj = Compact(U32).new()
        obj.encode(150000000000000)
        self.assertEqual(str(obj.data), "0x0b0060b7986c88")

    def test_compact_u32_encode_decode(self):

        value = 2000001

        obj = Compact(U32).new()
        data = obj.encode(value)

        obj = Compact(U32).new()

        self.assertEqual(obj.decode(data), value)

    def test_compact_u32_encode_decode_large(self):

        value = 2**30

        obj = Compact(U32).new()
        data = obj.encode(value)

        obj = Compact(U32).new()

        self.assertEqual(obj.decode(data), value)

    def test_compact_balance_encode_decode(self):
        scale_data = ScaleBytes('0x070010a5d4e8')
        value = 1000000000000

        Balance = U128

        obj = Compact(Balance).new()
        data = obj.encode(value)

        self.assertEqual(str(scale_data), str(data))

        self.assertEqual(obj.decode(data), value)

    def test_balance(self):
        Balance = U128
        obj = Compact(Balance).new()
        obj.decode(ScaleBytes("0x130080cd103d71bc22"))
        self.assertEqual(obj.value, 2503000000000000000)

    def test_compact_no_type(self):
        data = ScaleBytes("0x02093d00")
        compact = Compact().new()
        compact.decode(data)
        self.assertEqual(compact.value, 1000000)

        compact.encode(1000000)
        self.assertEqual(compact.data, data)


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.exceptions import ScaleEncodeException
from scalecodec.types import Array, U32, U8


class TestArray(unittest.TestCase):

    def test_dynamic_fixed_array_type_decode(self):
        obj = Array(U32, 1).new()
        self.assertEqual([1], obj.decode(ScaleBytes("0x01000000")))

        obj = Array(U32, 3).new()
        self.assertEqual([1, 2, 3], obj.decode(ScaleBytes("0x010000000200000003000000")))

        obj = Array(U32, 0).new()
        self.assertEqual([], obj.decode(ScaleBytes(bytes())))

    def test_dynamic_fixed_array_type_decode_u8(self):
        obj = Array(U8, 65).new()
        obj.decode(ScaleBytes(
                "0xc42b82d02bce3202f6a05d4b06d1ad46963d3be36fd0528bbe90e7f7a4e5fcd38d14234b1c9fcee920d76cfcf43b4ed5dd718e357c2bc1aae3a642975207e67f01"
        ))
        self.assertEqual(
            "0xc42b82d02bce3202f6a05d4b06d1ad46963d3be36fd0528bbe90e7f7a4e5fcd38d14234b1c9fcee920d76cfcf43b4ed5dd718e357c2bc1aae3a642975207e67f01",
            obj.value
        )

    def test_array_type_encode_u8(self):
        obj = Array(U8, 2).new()
        self.assertEqual('0x0102', str(obj.encode('0x0102')))
        self.assertEqual('0x0102', str(obj.encode(b'\x01\x02')))
        self.assertEqual('0x0102', str(obj.encode([1, 2])))

    def test_array_type_encode(self):
        obj = Array(U32, 2).new()
        self.assertEqual('0x0100000002000000', str(obj.encode([1, 2])))

        obj = Array(U8, 3).new()
        self.assertEqual('0x010203', str(obj.encode('0x010203')))

    def test_invalid_array_encode(self):
        obj = Array(U8, 3).new()
        self.assertRaises(ScaleEncodeException, obj.encode, '0x0102')

        obj = Array(U32, 3).new()
        self.assertRaises(ScaleEncodeException, obj.encode, '0x0102')

    def test_array_u8(self):
        obj = Array(U8, 4).new()

        value = [1, 2, 3, 4]
        data = obj.encode(value)

        self.assertEqual(data, ScaleBytes('0x01020304'))

        data.reset()
        self.assertEqual('0x01020304', obj.decode(data))

        self.assertEqual(obj.value_object, b'\x01\x02\x03\x04')


if __name__ == '__main__':
    unittest.main()



# Python SCALE Codec Library
#
# Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.types import BitVec


class TestBitvec(unittest.TestCase):

    def test_bitvec_decode(self):
        obj = BitVec().new()
        obj.decode(ScaleBytes('0x0c07'))
        self.assertEqual(obj.value, '0b111')

    def test_bitvec_decode_size2(self):
        obj = BitVec().new()
        obj.decode(ScaleBytes('0x0803'))
        self.assertEqual(obj.value, '0b11')

    def test_bitvec_decode_size_2bytes(self):
        obj = BitVec().new()
        obj.decode(ScaleBytes('0x28fd02'))
        self.assertEqual(obj.value, '0b1011111101')

    def test_bitvec_encode_list(self):
        obj = BitVec().new()
        data = obj.encode([True, True, True])
        self.assertEqual(data.to_hex(), '0x0c07')

    def test_bitvec_encode_list2(self):
        obj = BitVec().new()
        data = obj.encode([True, False])
        self.assertEqual(data.to_hex(), '0x0802')

    def test_bitvec_encode_list3(self):
        obj = BitVec().new()
        data = obj.encode([False, True])
        self.assertEqual(data.to_hex(), '0x0401')

    def test_bitvec_encode_list4(self):
        obj = BitVec().new()
        data = obj.encode([True, False, False, True, True, True, True, True, False, True])
        self.assertEqual(data.to_hex(), '0x287d02')

    def test_bitvec_encode_bin_str(self):
        obj = BitVec().new()
        data = obj.encode('0b00000111')
        self.assertEqual(data.to_hex(), '0x0c07')

    def test_bitvec_encode_bin_str2(self):
        obj = BitVec().new()
        data = obj.encode('0b00000010')
        self.assertEqual(data.to_hex(), '0x0802')

    def test_bitvec_encode_bin_str3(self):
        obj = BitVec().new()
        data = obj.encode('0b00000001')
        self.assertEqual(data.to_hex(), '0x0401')

    def test_bitvec_encode_bin_str4(self):
        obj = BitVec().new()
        data = obj.encode('0b00000010_01111101')
        self.assertEqual(data.to_hex(), '0x287d02')

    def test_bitvec_encode_int(self):
        obj = BitVec().new()
        data = obj.encode(0b00000111)
        self.assertEqual(data.to_hex(), '0x0c07')

    def test_bitvec_encode_int2(self):
        obj = BitVec().new()
        data = obj.encode(0b00000010)
        self.assertEqual(data.to_hex(), '0x0802')

    def test_bitvec_encode_int3(self):
        obj = BitVec().new()
        data = obj.encode(0b00000001)
        self.assertEqual(data.to_hex(), '0x0401')

    def test_bitvec_encode_int4(self):
        obj = BitVec().new()
        data = obj.encode(0b00000010_01111101)
        self.assertEqual(data.to_hex(), '0x287d02')

    def test_bitvec_encode_empty_list(self):
        obj = BitVec().new()
        data = obj.encode([])
        self.assertEqual(data.to_hex(), '0x00')


if __name__ == '__main__':
    unittest.main()



#  Polkascan API extension for Substrate Interface Library
#
#  Copyright 2018-2024 Stichting Polkascan (Polkascan Foundation).
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import unittest

from scalecodec.base import ScaleBytes
from scalecodec.exceptions import ScaleDeserializeException, ScaleEncodeException, ScaleDecodeException
from scalecodec.types import Struct, U8, Tuple, U32


class TestStruct(unittest.TestCase):

    def test_struct_encode(self):

        scale_obj = Struct(test=U8, test2=Tuple(U8, U8)).new()
        value = {'test': 2, "test2": (1, 5)}

        data = scale_obj.encode(value)

        self.assertEqual(ScaleBytes(data="0x020105"), data)

    def test_struct_encode2(self):

        obj = Struct(aye=U32, nay=U32).new()
        data = obj.encode({'aye': 4, 'nay': 2})

        self.assertEqual(ScaleBytes("0x0400000002000000"), data)

    def test_struct_encode_tuple(self):

        obj = Struct(aye=U32, nay=U32).new()
        data = obj.encode((4, 2))

        self.assertEqual(ScaleBytes("0x0400000002000000"), data)

    def test_struct_encode_int(self):

        obj = Struct(nonce=U32).new()
        data = obj.encode({'nonce': 1})

        self.assertEqual(ScaleBytes("0x01000000"), data)

    def test_struct_subclass(self):

        class Votes(Struct):
            arguments = {'aye': U32, 'nay': U32}

        obj = Votes().new()

        data = obj.encode({'aye': 4, 'nay': 2})

        self.assertEqual(ScaleBytes("0x0400000002000000"), data)

    def test_encode_missing_data(self):
        scale_obj = Struct(test=U8, test2=Tuple(U8, U8)).new()
        with self.assertRaises(ScaleEncodeException) as e:
            scale_obj.encode({'test': 2, "test3": (1, 5)})

    def test_decode(self):
        scale_obj = Struct(test=U8, test2=Tuple(U8, U8)).new()

        scale_obj.decode(ScaleBytes(data="0x020105"))
        self.assertDictEqual({'test': 2, "test2": (1, 5)}, scale_obj.value)

    def test_decode_remaining(self):
        scale_obj = Struct(test=U8, test2=Tuple(U8, U8)).new()

        with self.assertRaises(ScaleDecodeException) as e:
            scale_obj.decode(ScaleBytes(data="0x02010501"), check_remaining=True)

    def test_deserialize(self):
        scale_obj = Struct(test=U8, test2=Tuple(U8, U8)).new()
        scale_obj.deserialize({'test': 2, "test2": (1, 5)})

        self.assertIn('test', scale_obj.value_object)
        self.assertIn('test2', scale_obj.value_object)

    def test_deserialize_missing_data(self):
        scale_obj = Struct(test=U8, test2=Tuple(U8, U8)).new()
        with self.assertRaises(ScaleDeserializeException) as e:
            scale_obj.deserialize({'test': 2, "test3": (1, 5)})


if __name__ == '__main__':
    unittest.main()
